#include <avr/io.h>
#include <util/delay.h>
#include "ws2812.h"
#include "color.h"
#include "oled.h"
#include "i2c-bitbang.h"
#include "reset.h"

#define ROWS 5
#define COLS 6
/* 
 * Rows A-D: B0, B1, B2, B3
 * Cols 1-6: D5, D4, D3, D2, C2, B4
 * 
 * Diodes: ROW2COL.
 * 
 * Algorithm:
 * - set Rows as Input w/ Pullup 
 * - set Cols as tristate
 * - set one Col as Output, Lo
 * - read back Rows
 */

uint8_t is_left;

uint8_t state[COLS][ROWS];
LED_TYPE leds[COLS * ROWS];

void read_rows(uint8_t col) {
  if (is_left) {
    uint8_t row = PINB;

    for (uint8_t i = 0; i < ROWS; i++)
      state[col][i] = (state[col][0] << 1) | ((row >> i) & 1);
  } else {
    uint8_t row = ((PINC >> 7) & 0b1) | ((PIND >> 3 & 0b1111));

    for (uint8_t i = 0; i < ROWS; i++)
      state[col][i] = (state[col][0] << 1) | ((row >> i) & 1);
  }
}

void scan_keys() {
  if (is_left) {
    DDRB &= ~0b1111; // set rows as input
    PORTB |= 0b1111; // set pull-up

    DDRD |= 0b111100; // set Cols/D as output
    DDRC |= 0b100;    // set Cols/C as output
    DDRB |= 0b10000;  // set Cols/B as output
    PORTD |= 0b111100; // set Cols/D hi
    PORTC |= 0b100;    // set Cols/C hi
    PORTB |= 0b10000;  // set Cols/B hi

    // check d5
    PORTD &= ~(1 << 5);
    read_rows(0);
    PORTD |= 1 << 5;

    // check d4
    PORTD &= ~(1 << 4);
    read_rows(1);
    PORTD |= 1 << 4;

    // check d3
    PORTD &= ~(1 << 3);
    read_rows(2);
    PORTD |= 1 << 3;

    // check d2
    PORTD &= ~(1 << 2);
    read_rows(3);
    PORTD |= 1 << 2;

    // check c2
    PORTC &= ~(1 << 2);
    read_rows(4);
    PORTC |= 1 << 2;

    // check b4
    PORTB &= ~(1 << 4);
    read_rows(5);
    PORTB |= 1 << 4;
  } else {
    DDRC &= ~0b10000000; // set rows/C as input
    DDRD &= ~0b01111000; // set rows/D as input
    PORTC |= 0b10000000; // set rows/C pull-up
    PORTD |= 0b01111000; // set rows/D pull-up

    DDRB |= 0b11110001;  // set Cols as output
    PORTB |= 0b11110001;  // set Cols hi

    // check 1
    PORTB &= ~(1 << 0);
    read_rows(0);
    PORTB |= 1 << 0;

    // check 2
    PORTB &= ~(1 << 4);
    read_rows(1);
    PORTB |= 1 << 4;

    // check 3
    PORTB &= ~(1 << 5);
    read_rows(2);
    PORTB |= 1 << 5;

    // check 4
    PORTB &= ~(1 << 6);
    read_rows(2);
    PORTB |= 1 << 6;

    // check 5
    PORTB &= ~(1 << 7);
    read_rows(2);
    PORTB |= 1 << 7;
  }
}

uint8_t get_led_i(uint8_t c, uint8_t r) {
  if (c == 0) {
    c = 6;
  } else {
    c--;
  }

  return c + r * COLS;
}

void test_oled() {
  i2c_init();
  oled_init();
  oled_clear(0);
  oled_puts("Press any button...");
}

int main()
{
	CLKPR = (1 << CLKPCE); // enable a change to CLKPR
	CLKPR = 0; // set the CLKDIV to 0

	_delay_ms(200);

	test_oled();

  for (;;) {
    scan_keys();

    static uint8_t hue = 0;
    hue++;

    if ((state[3][0] & 0b1) == 0) {
      HSV l = { .h = 0, .s = 255, .v = 110 };
      leds[0] = hsv_to_rgb(l);
      l.h += 255 / 3;
      leds[1] = hsv_to_rgb(l);
      l.h += 255 / 3;
      leds[2] = hsv_to_rgb(l);
      ws2812_setleds(leds, COLS*ROWS);
      _delay_ms(500);
      reset_dfumode();
    }

    for (uint8_t c = 0; c < COLS; c++) {
      for (uint8_t r = 0; r < ROWS; r++) {
        uint8_t history = state[c][r];

        HSV l;
        
        l.h = hue;
        l.s = 255;
        l.v = 0;

        switch (history & 0b1) {
          case 0b01:
            l.v = 110;
            break;
          case 0b0:
            l.v = 0;
            break;
        }

        leds[get_led_i(c, r)] = hsv_to_rgb(l);
      }
    }
    ws2812_setleds(leds, COLS*ROWS);
    _delay_ms(5);
  }
}
